@@ -1,8 +1,7 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
 /*
- * Copyright (C) 2003 Imendio HB
- * Copyright (C) 2003 Mikael Hallendal <micke@imendio.com>
- * Copyright (C) 2003 CodeFactory AB. 
+ * Copyright (C) 2003-2004 Imendio HB
+ * Copyright (C) 2003-2004 Sjoerd Simons <sjoerd@luon.net>
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public License as
@@ -22,17 +21,16 @@
 
 #include <config.h>
 
-#ifdef HAVE_GNUTLS
-#include <gnutls/gnutls.h>
-#endif
-
 #include <string.h>
 #include <unistd.h>
 #include <sys/types.h>
+#include <fcntl.h>
+#include <errno.h>
 #ifndef __WIN32__
   #include <netdb.h>
   #include <sys/socket.h>
   #include <netinet/in.h>
+  #include <sys/time.h>
 #else
   #include <winsock2.h>
 #endif
@@ -53,24 +51,28 @@ typedef struct {
 } HandlerData;
 
 typedef struct {
-	GSource source;
+	LmConnection    *connection;
 	
+	/* struct to save resolved address */
+	struct addrinfo *resolved_addrs;
+	struct addrinfo *current_addr;
+	int              fd;
+	GIOChannel           *io_channel;
+} LmConnectData;
+
+typedef struct {
+	GSource       source;
 	LmConnection *connection;
 } LmIncomingSource;
 
 struct _LmConnection {
 	/* Parameters */
-	gchar          *server;
-	guint           port;
-	gboolean        use_ssl;
+	gchar      *server;
+	guint       port;
 
-#ifdef HAVE_GNUTLS
-	gnutls_session  gnutls_session;
-	gnutls_certificate_client_credentials gnutls_xcred;
-#endif
-	
-	gboolean    is_open;
-	gboolean    is_authenticated;
+	LmSSL      *ssl;
+
+	LmProxy    *proxy;
 	
 	LmParser   *parser;
 	gchar      *stream_id;
@@ -83,8 +85,12 @@ struct _LmConnection {
 	guint       io_watch_in;
 	guint       io_watch_err;
 	guint       io_watch_hup;
-
+	guint       fd;
+	
+	guint       open_id;
 	LmCallback *open_cb;
+
+	gboolean    cancel_open;
 	LmCallback *close_cb;
 	LmCallback *auth_cb;
 	LmCallback *register_cb;
@@ -94,6 +100,8 @@ struct _LmConnection {
 	LmQueue    *incoming_messages;
 	GSource    *incoming_source;
 
+	LmConnectionState state;
+
 	gint        ref_count;
 };
 
@@ -112,10 +120,10 @@ static void     connection_handle_message    (LmConnection         *connection,
 static void     connection_new_message_cb    (LmParser             *parser,
 					      LmMessage            *message,
 					      LmConnection         *connection);
-static gboolean connection_do_open         (LmConnection     *connection,
-					      GError          **error);
+static gboolean connection_do_open           (LmConnection         *connection,
+					      GError              **error);
 
-static void     connection_do_close           (LmConnection          *connection);
+static void     connection_do_close           (LmConnection        *connection);
 static gboolean connection_in_event          (GIOChannel   *source,
 					      GIOCondition  condition,
 					      LmConnection *connection);
@@ -157,9 +165,12 @@ static gboolean connection_incoming_check    (GSource         *source);
 static gboolean connection_incoming_dispatch (GSource         *source,
 					      GSourceFunc      callback,
 					      gpointer           user_data);
-static GSource * connection_create_source    (LmConnection *connection);
-static void      connection_signal_disconnect (LmConnection *connection,
-					       LmDisconnectReason reason);
+static GSource * connection_create_source       (LmConnection *connection);
+static void      connection_signal_disconnect   (LmConnection *connection,
+						 LmDisconnectReason reason);
+
+static void     connection_do_connect           (LmConnectData *connect_data);
+
 
 static GSourceFuncs incoming_funcs = {
 	connection_incoming_prepare,
@@ -190,8 +201,8 @@ connection_free (LmConnection *connection)
 	}
 
 	g_hash_table_destroy (connection->id_handlers);
-	
-	if (connection->is_open) {
+
+	if (lm_connection_is_open (connection)) {
 		connection_do_close (connection);
 	}
 
@@ -255,123 +266,58 @@ connection_new_message_cb (LmParser     *parser,
 }
 
 static gboolean
-connection_do_open (LmConnection *connection, GError **error)
+connection_succeeded (LmConnectData *connect_data)
 {
-	gint             fd = -1;
-	int              err = -1;
-	struct addrinfo  req;
-	struct addrinfo *ans;
-	struct addrinfo *tmpaddr;
-	char             name[NI_MAXHOST];
-	char             portname[NI_MAXSERV];
-	
-	g_return_val_if_fail (connection != NULL, FALSE);
-	
-	memset (&req, 0, sizeof(req));
+	LmConnection *connection = connect_data->connection;
+	LmMessage    *m;
+	GIOFlags      flags;
 
-	req.ai_family   = AF_UNSPEC;
-	req.ai_socktype = SOCK_STREAM;
-	req.ai_protocol = IPPROTO_TCP;
-
-	g_log (LM_LOG_DOMAIN,LM_LOG_LEVEL_NET,
-	       "Going to connect to %s\n",connection->server);
-	
-	if ((err = getaddrinfo (connection->server, NULL, &req, &ans)) != 0) {
-		g_set_error (error,
-			     LM_ERROR,                 
-			     LM_ERROR_CONNECTION_OPEN,   
-			     "getaddrinfo() failed");
+	/* Need some way to report error/success */
+	if (connection->cancel_open) {
 		return FALSE;
 	}
-
-#ifdef HAVE_GNUTLS
-	if (connection->use_ssl) {
-		gnutls_global_init ();
-		gnutls_certificate_allocate_credentials (&connection->gnutls_xcred);
-	}
-#endif
-
-	for (tmpaddr = ans ; tmpaddr != NULL ; tmpaddr = tmpaddr->ai_next) {
-		((struct sockaddr_in *) tmpaddr->ai_addr)->sin_port = htons (connection->port);
-
-		getnameinfo (tmpaddr->ai_addr,
-			     tmpaddr->ai_addrlen,
-			     name,     sizeof (name),
-			     portname, sizeof (portname),
-			     NI_NUMERICHOST | NI_NUMERICSERV);
-
-		g_log (LM_LOG_DOMAIN,LM_LOG_LEVEL_NET,
-		      "Trying %s port %s...\n", name, portname);
-
-		fd = socket (tmpaddr->ai_family, 
-			     tmpaddr->ai_socktype, 
-			     tmpaddr->ai_protocol);
-		if (fd < 0) {
-			continue;
-		}
-		
-		err = connect (fd,tmpaddr->ai_addr, tmpaddr->ai_addrlen);
-		if (err == 0) {
-			/* connection successfull */
-			break;
-		}
-		
-		close (fd);
-	}
 	
-	freeaddrinfo (ans);
+	connection->fd = connect_data->fd;
+	connection->io_channel = connect_data->io_channel;
 
-	/* check for failure */
-	if (fd < 0 || err < 0) {
-		g_set_error (error,
-			     LM_ERROR,
-			     LM_ERROR_CONNECTION_OPEN,
-			     "connection failed");
-		return FALSE;
-	}
+	freeaddrinfo (connect_data->resolved_addrs);
 
-#ifdef HAVE_GNUTLS
-	if (connection->use_ssl) {
-		int ret;
-		const int cert_type_priority[2] =
-		{ GNUTLS_CRT_X509, GNUTLS_CRT_OPENPGP };
+	/* don't need this anymore */
+	g_free(connect_data);
 
-		gnutls_init (&connection->gnutls_session, GNUTLS_CLIENT);
-		gnutls_set_default_priority (connection->gnutls_session);
-		gnutls_certificate_type_set_priority (connection->gnutls_session,
-						      cert_type_priority);
-		gnutls_credentials_set (connection->gnutls_session,
-					GNUTLS_CRD_CERTIFICATE,
-					connection->gnutls_xcred);
-		
-		gnutls_transport_set_ptr (connection->gnutls_session, 
-					  (gnutls_transport_ptr) fd);
+	flags = g_io_channel_get_flags (connection->io_channel);
 
-		ret = gnutls_handshake (connection->gnutls_session);
-		
-		if (ret < 0) {
-			gnutls_perror (ret);
-			shutdown (fd, SHUT_RDWR);
-			close (fd);
+	/* unset the nonblock flag */
+	flags &= ~G_IO_FLAG_NONBLOCK;
+	
+	/* unset the nonblocking stuff for some time, because GNUTLS doesn't 
+	 * like that */
+	g_io_channel_set_flags (connection->io_channel, flags, NULL);
+
+	if (connection->ssl) {
+		if (!_lm_ssl_begin (connection->ssl, connection->fd, 
+				    connection->server, NULL)) {
+			shutdown (connection->fd, SHUT_RDWR);
+			close (connection->fd);
 			connection_do_close (connection);
-			g_set_error (error, LM_ERROR, LM_ERROR_CONNECTION_OPEN,
-				     "*** GNUTLS handshake failed");
+			connection->fd = -1;
+			g_io_channel_unref(connection->io_channel);
 			return FALSE;
 		}
 	}
-#endif
 	
-	connection->io_channel = g_io_channel_unix_new (fd);
 	g_io_channel_set_close_on_unref (connection->io_channel, TRUE);
 	g_io_channel_set_encoding (connection->io_channel, NULL, NULL);
 	
 	g_io_channel_set_buffered (connection->io_channel, FALSE);
 	g_io_channel_set_flags (connection->io_channel,
-				G_IO_FLAG_NONBLOCK, NULL);
+				flags & G_IO_FLAG_NONBLOCK, NULL);
+	
 	connection->io_watch_in = g_io_add_watch (connection->io_channel,
 						  G_IO_IN,
 						  (GIOFunc) connection_in_event,
 						  connection);
+	
 	connection->io_watch_err = g_io_add_watch (connection->io_channel, 
 						   G_IO_ERR,
 						   (GIOFunc) connection_error_event,
@@ -381,17 +327,241 @@ connection_do_open (LmConnection *connection, GError **error)
 						   (GIOFunc) connection_hup_event,
 						   connection);
 
-	connection->is_open = TRUE;
-
-	if (!connection_send (connection,
-			      "<?xml version='1.0' encoding='UTF-8'?>", -1, 
-			      error)) {
+	if (!connection_send (connection, 
+			      "<?xml version='1.0' encoding='UTF-8'?>", -1,
+			      NULL)) {
+		connection_do_close (connection);
 		return FALSE;
 	}
 
-	return TRUE;
+	m = lm_message_new (connection->server, LM_MESSAGE_TYPE_STREAM);
+	lm_message_node_set_attributes (m->node,
+					"xmlns:stream", 
+					"http://etherx.jabber.org/streams",
+					"xmlns", "jabber:client",
+					NULL);
+	
+	lm_verbose ("Opening stream...");
+	
+	if (!lm_connection_send (connection, m, NULL)) {
+		lm_message_unref (m);
+		connection_do_close (connection);
+		return FALSE;
+	}
+		
+	lm_message_unref (m);
+
+	/* Success */
+	return FALSE;
 }
 
+static void 
+connection_failed_with_error (LmConnectData *connect_data, int error) 
+{
+	g_log (LM_LOG_DOMAIN, LM_LOG_LEVEL_NET,
+	       "Connection failed: %s (error %d)\n",
+	       strerror (error), error);
+	
+	connect_data->current_addr = connect_data->current_addr->ai_next;
+	
+	if (connect_data->io_channel != NULL) {
+		g_io_channel_unref (connect_data->io_channel);
+	}
+	
+	if (connect_data->current_addr == NULL) {
+		freeaddrinfo (connect_data->resolved_addrs);
+		g_free (connect_data);
+	} else {
+		/* try to connect to the next host */
+		connection_do_connect (connect_data);
+	}
+}
+
+static void 
+connection_failed (LmConnectData *connect_data)
+{
+	connection_failed_with_error (connect_data,errno);
+}
+	
+static gboolean 
+connection_connect_cb (GIOChannel   *source, 
+		       GIOCondition  condition,
+		       gpointer      data) 
+{
+	LmConnection  *connection;
+	LmConnectData *connect_data;
+	int            error;
+	int            len  = sizeof(error);
+
+	connect_data = (LmConnectData *) data;
+	connection   = connect_data->connection;
+	
+	if (condition == G_IO_ERR) {
+		/* get the real error from the socket */
+		getsockopt (connect_data->fd, SOL_SOCKET, SO_ERROR, 
+			    &error, &len);
+		connection_failed_with_error (connect_data, error);
+		return FALSE;
+	} else if (condition == G_IO_OUT) {
+		if (connect_data->connection->proxy) {
+			if (!_lm_proxy_negotiate (connection->proxy, connect_data->fd, connection->server, connection->port)) {
+				connection_failed (connect_data);
+				return FALSE;
+			}
+		}
+		connection_succeeded (connect_data);
+	} else {
+		g_assert_not_reached ();
+	}
+
+	return FALSE;
+}
+
+static void
+connection_do_connect (LmConnectData *connect_data) 
+{
+	LmConnection    *connection;
+	int              fd;
+	int              res;
+	int              port;
+	int              flags;
+	char             name[NI_MAXHOST];
+	char             portname[NI_MAXSERV];
+	struct addrinfo *addr;
+	
+	connection = connect_data->connection;
+	addr = connect_data->current_addr;
+ 
+	if (connection->proxy) {
+		port = htons (lm_proxy_get_port (connection->proxy));
+	} else {
+		port = htons (connection->port);
+	}
+	
+	((struct sockaddr_in *) addr->ai_addr)->sin_port = port;
+
+	getnameinfo (addr->ai_addr,
+		     addr->ai_addrlen,
+		     name,     sizeof (name),
+		     portname, sizeof (portname),
+		     NI_NUMERICHOST | NI_NUMERICSERV);
+
+	g_log (LM_LOG_DOMAIN,LM_LOG_LEVEL_NET,
+	       "Trying %s port %s...\n", name, portname);
+	
+	fd = socket (addr->ai_family, 
+		     addr->ai_socktype, 
+		     addr->ai_protocol);
+	
+	if (fd < 0) {
+		connection_failed (connect_data);
+		return;
+	}
+
+	flags = fcntl (fd, F_GETFL, 0);
+	fcntl (fd, F_SETFL, flags | O_NONBLOCK);
+	
+	res = connect (fd, addr->ai_addr, addr->ai_addrlen);
+	connect_data->fd = fd;
+	
+	if (res < 0 && errno != EINPROGRESS) {
+		close (fd);
+		connection_failed (connect_data);
+		return;
+	}
+	
+	connect_data->io_channel = g_io_channel_unix_new (fd);
+	g_io_add_watch (connect_data->io_channel, G_IO_OUT|G_IO_ERR,
+			(GIOFunc) connection_connect_cb, connect_data);
+	return;
+}
+
+/* Returns directly */
+/* Setups all data needed to start the connection attempts */
+static gboolean
+connection_do_open (LmConnection *connection, GError **error) 
+{
+	struct addrinfo  req;
+	struct addrinfo *ans;
+	LmConnectData   *data;
+
+	if (lm_connection_is_open (connection)) {
+		g_set_error (error,
+			     LM_ERROR,
+			     LM_ERROR_CONNECTION_NOT_OPEN,
+			     "Connection is already open, call lm_connection_close() first");
+		return FALSE;
+	}
+
+	if (!connection->server) {
+		g_set_error (error,
+			     LM_ERROR,
+			     LM_ERROR_CONNECTION_OPEN,
+			     "You need to set the server hostname in the call to lm_connection_new()");
+		return FALSE;
+	}
+
+	/* source thingie for messages and stuff */
+	connection->incoming_source = connection_create_source (connection);
+	g_source_attach (connection->incoming_source,NULL);
+	
+	lm_verbose ("Connecting to: %s:%d\n", 
+		    connection->server, connection->port);
+
+	memset (&req, 0, sizeof(req));
+
+	req.ai_family   = AF_UNSPEC;
+	req.ai_socktype = SOCK_STREAM;
+	req.ai_protocol = IPPROTO_TCP;
+	
+	connection->cancel_open = FALSE;
+	connection->state = LM_CONNECTION_STATE_CONNECTING;
+	
+	if (connection->proxy) {
+		const gchar *proxy_server;
+
+		proxy_server = lm_proxy_get_server (connection->proxy);
+		/* connect through proxy */
+		g_log (LM_LOG_DOMAIN,LM_LOG_LEVEL_NET,
+		       "Going to connect to %s\n", proxy_server);
+
+		if (getaddrinfo (proxy_server, NULL, &req, &ans) != 0) {
+			g_set_error (error,
+				     LM_ERROR,                 
+				     LM_ERROR_CONNECTION_OPEN,   
+				     "getaddrinfo() failed");
+			return FALSE;
+		}
+	} else { /* connect directly */
+		g_log (LM_LOG_DOMAIN,LM_LOG_LEVEL_NET,
+		       "Going to connect to %s\n",connection->server);
+
+		if (getaddrinfo (connection->server, NULL, &req, &ans) != 0) {
+			g_set_error (error,
+				     LM_ERROR,                 
+				     LM_ERROR_CONNECTION_OPEN,   
+				     "getaddrinfo() failed");
+			return FALSE;
+		}
+	}
+
+	if (connection->ssl) {
+		_lm_ssl_initialize (connection->ssl);
+	}
+
+	/* Prepare and do the nonblocking connection */
+	data = g_new (LmConnectData, 1);
+
+	data->connection     = connection;
+	data->resolved_addrs = ans;
+	data->current_addr   = ans;
+	data->io_channel     = NULL;
+	data->fd             = -1;
+
+	connection_do_connect (data);
+	return TRUE;
+}
+					
 static void
 connection_do_close (LmConnection *connection)
 {
@@ -402,21 +572,20 @@ connection_do_close (LmConnection *connection)
 
 		g_io_channel_unref (connection->io_channel);
 		connection->io_channel = NULL;
-	} 
+	}
 
-	if (!connection->is_open) {
+	g_source_remove (g_source_get_id (connection->incoming_source));
+	g_source_unref (connection->incoming_source);
+
+	if (!lm_connection_is_open (connection)) {
 		return;
 	}
 
-	connection->is_open = FALSE;
+	connection->state = LM_CONNECTION_STATE_DISCONNECTED;
 
-#ifdef HAVE_GNUTLS
-	if (connection->use_ssl) {
-		gnutls_deinit (connection->gnutls_session);
-		gnutls_certificate_free_credentials (connection->gnutls_xcred);
-		gnutls_global_deinit ();
+	if (connection->ssl) {
+		_lm_ssl_close (connection->ssl);
 	}
-#endif
 }
 
 
@@ -433,32 +602,15 @@ connection_in_event (GIOChannel   *source,
 		return FALSE;
 	}
 
-#ifdef HAVE_GNUTLS
-	if (connection->use_ssl) {
-		bytes_read = gnutls_record_recv (connection->gnutls_session,
-						 buf,IN_BUFFER_SIZE - 1);
-		if (bytes_read == GNUTLS_E_AGAIN) {
-			status = G_IO_STATUS_AGAIN;
-		}
-		else if (bytes_read <= 0) {
-			status = G_IO_STATUS_ERROR;
-			
-			//connection_error_event (connection->io_channel, 
-			//			G_IO_HUP,
-			//			connection);
-		}
-		else {
-			status = G_IO_STATUS_NORMAL;
-		}
+	if (connection->ssl) {
+		status = _lm_ssl_read (connection->ssl, 
+				       buf, IN_BUFFER_SIZE - 1, &bytes_read);
 	} else {
-#endif
-	    status = g_io_channel_read_chars (connection->io_channel,
-					      buf, IN_BUFFER_SIZE - 1,
-					      &bytes_read,
-					      NULL);
-#ifdef HAVE_GNUTLS
+		status = g_io_channel_read_chars (connection->io_channel,
+						  buf, IN_BUFFER_SIZE - 1,
+						  &bytes_read,
+						  NULL);
 	}
-#endif
 
 	if (status != G_IO_STATUS_NORMAL) {
 		gint reason;
@@ -484,7 +636,8 @@ connection_in_event (GIOChannel   *source,
 	}
 
 	buf[bytes_read] = '\0';
-	g_log (LM_LOG_DOMAIN, LM_LOG_LEVEL_NET, "\nRECV:\n");
+	g_log (LM_LOG_DOMAIN, LM_LOG_LEVEL_NET, "\nRECV [%d]:\n", 
+	       bytes_read);
 	g_log (LM_LOG_DOMAIN, LM_LOG_LEVEL_NET, 
 	       "-----------------------------------\n");
 	g_log (LM_LOG_DOMAIN, LM_LOG_LEVEL_NET, "'%s'\n", buf);
@@ -540,7 +693,7 @@ connection_send (LmConnection  *connection,
 {
 	gsize             bytes_written;
 	
-	if (!lm_connection_is_open (connection)) {
+	if (connection->state < LM_CONNECTION_STATE_CONNECTING) {
 		g_set_error (error,
 			     LM_ERROR,
 			     LM_ERROR_CONNECTION_NOT_OPEN,
@@ -558,24 +711,18 @@ connection_send (LmConnection  *connection,
 	g_log (LM_LOG_DOMAIN, LM_LOG_LEVEL_NET, "%s\n", str);
 	g_log (LM_LOG_DOMAIN, LM_LOG_LEVEL_NET, 
 	       "-----------------------------------\n");
-	
-#ifdef HAVE_GNUTLS
-	if (connection->use_ssl) {
-		while ((bytes_written = gnutls_record_send (connection->gnutls_session, str, len)) < 0)
-			if (bytes_written != GNUTLS_E_INTERRUPTED &&
-			    bytes_written != GNUTLS_E_AGAIN)
-			{
-				connection_error_event (connection->io_channel, G_IO_HUP,
-							connection);
-			}
-		    
+
+	if (connection->ssl) {
+		if (!_lm_ssl_send (connection->ssl, str, len)) {
+			
+			connection_error_event (connection->io_channel, 
+						G_IO_HUP,
+						connection);
+		}
 	} else {
-#endif
 		g_io_channel_write_chars (connection->io_channel, str, len, 
 					  &bytes_written, NULL);
-#ifdef HAVE_GNUTLS
 	}
-#endif
 
 	return TRUE;
 }
@@ -733,9 +880,11 @@ connection_auth_reply (LmMessageHandler *handler,
 	type = lm_message_node_get_attribute (m->node, "type");
 	if (strcmp (type, "result") == 0) {
 		result = TRUE;
+		connection->state = LM_CONNECTION_STATE_AUTHENTICATED;
 	} 
 	else if (strcmp (type, "error") == 0) {
 		result = FALSE;
+		connection->state = LM_CONNECTION_STATE_CONNECTED;
 	}
 	
 	lm_verbose ("AUTH reply: %d\n", result);
@@ -767,6 +916,8 @@ connection_stream_received (LmConnection *connection, LmMessage *m)
 	
 	lm_verbose ("Stream received: %s\n", connection->stream_id);
 	
+	connection->state = LM_CONNECTION_STATE_CONNECTED;
+	
 	/* Check to see if the stream is correctly set up */
 	result = TRUE;
 
@@ -873,17 +1024,18 @@ lm_connection_new (const gchar *server)
 	}
 	
 	connection->port              = LM_CONNECTION_DEFAULT_PORT;
-	connection->use_ssl           = FALSE;
+	connection->ssl               = NULL;
+	connection->proxy             = NULL;
 	connection->disconnect_cb     = NULL;
 	connection->incoming_messages = lm_queue_new ();
-	connection->incoming_source   = connection_create_source (connection);
+	connection->cancel_open       = FALSE;
+	connection->state             = LM_CONNECTION_STATE_DISCONNECTED;
 	
 	connection->id_handlers = g_hash_table_new_full (g_str_hash, 
 							 g_str_equal,
 							 g_free, 
 							 (GDestroyNotify) lm_message_handler_unref);
 	connection->ref_count         = 1;
-	g_source_attach (connection->incoming_source, NULL);
 	
 	for (i = 0; i < LM_MESSAGE_TYPE_UNKNOWN; ++i) {
 		connection->handlers[i] = NULL;
@@ -896,6 +1048,7 @@ lm_connection_new (const gchar *server)
 	return connection;
 }
 
+
 /**
  * lm_connection_open:
  * @connection: #LmConnection to open
@@ -915,53 +1068,17 @@ lm_connection_open (LmConnection      *connection,
 		    GDestroyNotify     notify,
 		    GError           **error)
 {
-	LmMessage *m;
-	gboolean   result;
-	
 	g_return_val_if_fail (connection != NULL, FALSE);
 	
-	if (lm_connection_is_open (connection)) {
-		g_set_error (error,
-			     LM_ERROR,
-			     LM_ERROR_CONNECTION_NOT_OPEN,
-			     "Connection is already open, call lm_connection_close() first");
-		return FALSE;
-	}
+	connection->open_cb = _lm_utils_new_callback (function, 
+						      user_data, notify);
 
-	if (!connection->server) {
-		g_set_error (error,
-			     LM_ERROR,
-			     LM_ERROR_CONNECTION_OPEN,
-			     "You need to set the server hostname in the call to lm_connection_new()");
-		return FALSE;
-	}
-
-	connection->open_cb = _lm_utils_new_callback (function, user_data, notify);
-	
-	lm_verbose ("Connecting to: %s:%d\n", 
-		    connection->server, connection->port);
-	
-	if (!connection_do_open (connection, error)) {
-		return FALSE;
-	}
-	
-	m = lm_message_new (connection->server, LM_MESSAGE_TYPE_STREAM);
-	lm_message_node_set_attributes (m->node,
-					"xmlns:stream", "http://etherx.jabber.org/streams",
-					"xmlns", "jabber:client",
-					NULL);
-	
-	lm_verbose ("Opening stream...");
-	
-	result = lm_connection_send (connection, m, error);
-	lm_message_unref (m);
-	
-	return result;
+	return connection_do_open (connection, error);
 }
 
 /**
  * lm_connection_open_and_block:
- * @connection: an #LmConnection
+ * @connection: an #LmConnection to open
  * @error: location to store error, or %NULL
  * 
  * Opens @connection and waits until the stream is setup. 
@@ -971,78 +1088,49 @@ lm_connection_open (LmConnection      *connection,
 gboolean
 lm_connection_open_and_block (LmConnection *connection, GError **error)
 {
-	LmMessage *m;
-	gboolean   result;
-	gboolean   finished = FALSE;
-	gboolean   ret_val = FALSE;
-
+	gboolean          result;
+	LmConnectionState state;
+	
 	g_return_val_if_fail (connection != NULL, FALSE);
+
+	connection->open_cb = NULL;
+	result = connection_do_open (connection, error);
+
+	if (result == FALSE) {
+		return FALSE;
+	}
+	
+	while ((state = lm_connection_get_state (connection)) == LM_CONNECTION_STATE_CONNECTING) {
+		g_print ("while...\n");
+		if (g_main_context_pending (NULL)) {
+			g_print ("Iterating\n");
+			g_main_context_iteration (NULL, TRUE);
+		} else {
+			usleep (10);
+		}
+	}
+
+	g_print ("Connected!\n");
 	
 	if (lm_connection_is_open (connection)) {
-		g_set_error (error,
-			     LM_ERROR,
-			     LM_ERROR_CONNECTION_NOT_OPEN,
-			     "Connection is already open, call lm_connection_close() first");
-		return FALSE;
-	}
-	if (!connection->server) {
-		g_set_error (error,
-			     LM_ERROR,
-			     LM_ERROR_CONNECTION_OPEN,
-			     "You need to set the server hostname in the call to lm_connection_new()");
-		return FALSE;
+		return TRUE;
 	}
 
-	lm_verbose ("(Block)Connecting to: %s:%d\n", 
-		    connection->server, connection->port);
-	
-	if (!connection_do_open (connection, error)) {
-		return FALSE;
-	}
-	
-	m = lm_message_new (connection->server, LM_MESSAGE_TYPE_STREAM);
-	lm_message_node_set_attributes (m->node,
-					"xmlns:stream", "http://etherx.jabber.org/streams",
-					"xmlns", "jabber:client",
-					NULL);
-	
-	lm_verbose ("Sending stream: \n%s\n", 
-		    lm_message_node_to_string (m->node));
-	
-	result = lm_connection_send (connection, m, error);
-	lm_message_unref (m);
+	return FALSE;
+}
 
- 	g_source_remove (g_source_get_id (connection->incoming_source));
-	g_source_unref (connection->incoming_source);
+/**
+ * lm_connection_cancel_open:
+ * @connection: an #LmConnection to cancel opening on
+ *
+ * Cancels the open operation of a connection. The connection should be in the state #LM_CONNECTION_STATE_CONNECTING.
+ **/
+void
+lm_connection_cancel_open (LmConnection *connection)
+{
+	g_return_if_fail (connection != NULL);
 
-	while (!finished) {
-		gint n;
-		
-		g_main_context_iteration (NULL, TRUE);
-		
-		if (lm_queue_is_empty (connection->incoming_messages)) {
-			continue;
-		}
-
-		for (n = 0; n < connection->incoming_messages->length; n++) {
-			LmMessage *m;
-
-			m = lm_queue_peek_nth (connection->incoming_messages, n);
-			if (lm_message_get_type (m) == LM_MESSAGE_TYPE_STREAM) {
-				connection->stream_id = 
-					g_strdup (lm_message_node_get_attribute (m->node, "id"));
-				ret_val = TRUE;
-				finished = TRUE;
-				lm_queue_remove_nth (connection->incoming_messages, n);
-				break;
-			}
-		}
-	}
-	
-	connection->incoming_source = connection_create_source (connection);
-	g_source_attach (connection->incoming_source, NULL);
-
-	return ret_val;
+	connection->cancel_open = TRUE;
 }
 
 /**
@@ -1050,7 +1138,7 @@ lm_connection_open_and_block (LmConnection *connection, GError **error)
  * @connection: #LmConnection to close 
  * @error: location to store error, or %NULL
  * 
- * A synchronos call to close the connection. When returning the connection is considered to be closed and can be opened again with lm_connection_open().
+ * A synchronous call to close the connection. When returning the connection is considered to be closed and can be opened again with lm_connection_open().
  * 
  * Return value: Returns #TRUE if no errors where detected, otherwise #FALSE.
  **/
@@ -1095,7 +1183,7 @@ lm_connection_close (LmConnection      *connection,
  * @notify: Destroy function to free the memory used by @user_data, can be NULL.
  * @error: location to store error, or %NULL
  * 
- * Tries to authenticate a user against the server. The #LmResult in the result callback will tell if it succeeded or not. 
+ * Tries to authenticate a user against the server. The #LmResult in the result callback @function will say whether it succeeded or not. 
  * 
  * Return value: #TRUE if no errors where detected while sending the authentication message, #FALSE otherwise.
  **/
@@ -1126,6 +1214,8 @@ lm_connection_authenticate (LmConnection      *connection,
 			     "Connection is not open, call lm_connection_open() first");
 		return FALSE;
 	}
+
+	connection->state = LM_CONNECTION_STATE_AUTHENTICATING;
 	
 	connection->auth_cb = _lm_utils_new_callback (function, 
 						      user_data, 
@@ -1193,6 +1283,7 @@ lm_connection_authenticate_and_block (LmConnection  *connection,
 		return FALSE;
 	}
 
+	g_print ("Foo: %d\n", (gint)connection);
 	m = connection_create_auth_msg (connection,
 					username, 
 					password,
@@ -1213,6 +1304,10 @@ lm_connection_authenticate_and_block (LmConnection  *connection,
 		return TRUE;
 		break;
 	case LM_MESSAGE_SUB_TYPE_ERROR:
+		g_set_error (error,
+			     LM_ERROR,
+			     LM_ERROR_AUTH_FAILED,
+			     "Authentication failed");
 		return FALSE;
 		break;
 	default:
@@ -1234,7 +1329,9 @@ lm_connection_authenticate_and_block (LmConnection  *connection,
 gboolean
 lm_connection_is_open (LmConnection *connection)
 {
-	return connection->is_open;
+	g_return_val_if_fail (connection != NULL, FALSE);
+	
+	return connection->state >= LM_CONNECTION_STATE_CONNECTED;
 }
 
 /**
@@ -1248,7 +1345,9 @@ lm_connection_is_open (LmConnection *connection)
 gboolean 
 lm_connection_is_authenticated (LmConnection *connection)
 {
-	return connection->is_authenticated;
+	g_return_val_if_fail (connection != NULL, FALSE);
+
+	return connection->state >= LM_CONNECTION_STATE_AUTHENTICATED;
 }
 
 /**
@@ -1262,6 +1361,8 @@ lm_connection_is_authenticated (LmConnection *connection)
 const gchar *
 lm_connection_get_server (LmConnection *connection)
 {
+	g_return_val_if_fail (connection != NULL, NULL);
+
 	return connection->server;
 }
 
@@ -1270,11 +1371,14 @@ lm_connection_get_server (LmConnection *connection)
  * @connection: an #LmConnection
  * @server: Address of the server
  * 
- * Sets the server address to @connection. Notice that @connection can't be open while doing this.
+ * Sets the server address for @connection to @server. Notice that @connection can't be open while doing this.
  **/
 void
 lm_connection_set_server (LmConnection *connection, const gchar *server)
 {
+	g_return_if_fail (connection != NULL);
+	g_return_if_fail (server != NULL);
+	
 	if (lm_connection_is_open (connection)) {
 		g_warning ("Can't change server address while connected");
 		return;
@@ -1291,13 +1395,15 @@ lm_connection_set_server (LmConnection *connection, const gchar *server)
  * lm_connection_get_port:
  * @connection: an #LmConnection
  * 
- * Fetches the port tnat @connection is using.
+ * Fetches the port that @connection is using.
  * 
  * Return value: 
  **/
 guint
 lm_connection_get_port (LmConnection *connection)
 {
+	g_return_val_if_fail (connection != NULL, 0);
+
 	return connection->port;
 }
 
@@ -1306,11 +1412,13 @@ lm_connection_get_port (LmConnection *connection)
  * @connection: an #LmConnection
  * @port: server port
  * 
- * Sets the server port of that @connection will be using.
+ * Sets the server port that @connection will be using.
  **/
 void
 lm_connection_set_port (LmConnection *connection, guint port)
 {
+	g_return_if_fail (connection != NULL);
+	
 	if (lm_connection_is_open (connection)) {
 		g_warning ("Can't change server port while connected");
 		return;
@@ -1320,57 +1428,92 @@ lm_connection_set_port (LmConnection *connection, guint port)
 }
 
 /**
- * lm_connection_supports_ssl:
+ * lm_connection_get_ssl: 
+ * @connection: an #LmConnection
  *
- * Checks whether Loudmouth supports SSL or not
- *
- * Return value: #TRUE if this installation of Loudmouth supports SSL, otherwise returnes #FALSE.
+ * Returns the SSL struct if the connection is using one.
+ * 
+ * Return value: The ssl struct or %NULL if no proxy is used.
  **/
-gboolean
-lm_connection_supports_ssl (void)
+LmSSL *
+lm_connection_get_ssl (LmConnection *connection)
 {
-#ifdef HAVE_GNUTLS
-	return TRUE;
-#else
-	return FALSE;
-#endif
+	g_return_val_if_fail (connection != NULL, NULL);
+	
+	return connection->ssl;
 }
 
 /**
- * lm_connection_get_use_ssl:
- * @connection: an #LmConnection
- * 
- * Fetches if @connection is using SSL or not
- * 
- * Return value: #TRUE if @connection is using SSL, #FALSE otherwise.
- **/
-gboolean
-lm_connection_get_use_ssl (LmConnection *connection)
-{
-	return connection->use_ssl;
-}
-
-/**
- * lm_connection_set_use_ssl:
- * @connection: an #LmConnection
- * @use_ssl: whether to use SSL or not.
- * 
- * Sets whether @connection should use SSL for encryping traffic to/from the server.
- **/
+ * lm_connection_set_ssl:
+ * @connection: An #LmConnection
+ * @ssl: An #LmSSL
+ *
+ * Sets SSL struct or unset if @ssl is %NULL. If set @connection will use SSL to for the connection.
+ */
 void
-lm_connection_set_use_ssl (LmConnection *connection, gboolean use_ssl)
+lm_connection_set_ssl (LmConnection *connection, LmSSL *ssl)
 {
-	if (lm_connection_is_open (connection)) {
-		g_warning ("use_ssl can't be changed while connected");
-		return;
+	g_return_if_fail (connection != NULL);
+
+	if (connection->ssl) {
+		lm_ssl_unref (connection->ssl);
 	}
 
-	connection->use_ssl = use_ssl;
+	if (ssl) {
+		connection->ssl = lm_ssl_ref (ssl);
+	} else {
+		connection->ssl = NULL;
+	}
 }
 
+#if 0
+/**
+ * lm_connection_get_proxy: 
+ * @connection: an #LmConnection
+ *
+ * Returns the proxy if the connection is using one.
+ * 
+ * Return value: The proxy or %NULL if no proxy is used.
+ **/
+LmProxy *
+lm_connection_get_proxy (LmConnection *connection)
+{
+	g_return_val_if_fail (connection != NULL, NULL);
+
+	return connection->proxy;
+} 
+
+/**
+ * lm_connection_set_proxy: 
+ * @connection: an #LmConnection
+ * @proxy: an #LmProxy
+ *
+ * Sets the proxy to use for this connection.
+ * 
+ * Return value: The proxy or %NULL if no proxy is used. Notice that @connection can't be open while doing this.
+ **/
+void
+lm_connection_set_proxy (LmConnection *connection, LmProxy *proxy)
+{
+	g_return_if_fail (connection != NULL);
+	g_return_if_fail (proxy != NULL);
+
+	if (lm_connection_is_open (connection)) {
+		g_warning ("Can't change server proxy while connected");
+		return;
+	}
+	
+	if (connection->proxy) {
+		lm_proxy_unref (connection->proxy);
+	}
+
+	connection->proxy = lm_proxy_ref (proxy);
+}
+#endif
+
 /**
  * lm_connection_send: 
- * @connection: #LmConnection to send connection over.
+ * @connection: #LmConnection to send message over.
  * @message: #LmMessage to send.
  * @error: location to store error, or %NULL
  * 
@@ -1410,7 +1553,7 @@ lm_connection_send (LmConnection  *connection,
  * 
  * Send a #LmMessage which will result in a reply. 
  * 
- * Return value: 
+ * Return value: Returns #TRUE if no errors where detected while sending, #FALSE otherwise.
  **/
 gboolean 
 lm_connection_send_with_reply (LmConnection      *connection,
@@ -1456,6 +1599,9 @@ lm_connection_send_with_reply_and_block (LmConnection  *connection,
 	gchar     *id;
 	LmMessage *reply = NULL;
 
+	g_return_val_if_fail (connection != NULL, NULL);
+	g_return_val_if_fail (message != NULL, NULL);
+
 	if (lm_message_node_get_attribute (message->node, "id")) {
 		id = g_strdup (lm_message_node_get_attribute (message->node, 
 							      "id"));
@@ -1586,6 +1732,8 @@ lm_connection_set_disconnect_function (LmConnection         *connection,
 				       gpointer              user_data,
 				       GDestroyNotify        notify)
 {
+	g_return_if_fail (connection != NULL);
+
 	if (connection->disconnect_cb) {
 		_lm_utils_free_callback (connection->disconnect_cb);
 	}
@@ -1615,6 +1763,22 @@ lm_connection_send_raw (LmConnection  *connection,
 
 	return connection_send (connection, str, -1, error);
 }
+/**
+ * lm_connection_get_state:
+ * @connection: Connection to get state on
+ *
+ * Returns the state of the connection.
+ *
+ * Return value: The state of the connection.
+ **/
+LmConnectionState 
+lm_connection_get_state (LmConnection *connection)
+{
+	g_return_val_if_fail (connection != NULL, 
+			      LM_CONNECTION_STATE_DISCONNECTED);
+
+	return connection->state;
+}
 
 /**
  * lm_connection_ref:
@@ -1640,7 +1804,7 @@ lm_connection_ref (LmConnection *connection)
  * @connection: Connection to remove reference from.
  * 
  * Removes a reference on @connection. If there are no references to
- * @connection it will be fried and shouldn't be used again.
+ * @connection it will be freed and shouldn't be used again.
  **/
 void
 lm_connection_unref (LmConnection *connection)
@@ -1654,45 +1818,3 @@ lm_connection_unref (LmConnection *connection)
 	}
 }
 
-#if 0
-void
-lm_connection_register (LmConnection           *connection,
-		    const gchar        *username,
-		    const gchar        *password,
-		    const gchar        *resource,
-		    LmRegisterCallback  callback,
-		    gpointer            user_data)
-{
-	LmElement    *element;
-	LmNode       *q_node;
-	gchar        *id;
-	static gint   register_id = 0;
-	
-	g_return_if_fail (connection != NULL);
-	g_return_if_fail (lm_connection_is_open (connection));
-	
-	/* Use lm:iq:register name space */
-
-	element = lm_iq_new (LM_IQ_TYPE_SET);
-	
-	q_node = lm_node_new ("query");
-	lm_node_set_attribute (q_node, "xmlns", JABBER_IQ_REGISTER);
-	lm_node_add_child (q_node, "username", username);
-	lm_node_add_child (q_node, "password", password);
-	lm_node_add_child (q_node, "resource", resource);
-
-	lm_element_add_child_node (element, q_node);
-	
-	id = g_strdup_printf ("register_%d", ++register_id);
-	lm_element_set_id (element, id);
-	
-	lm_connection_send (connection, element, NULL);
-	connection_add_callback (connection, id, callback, user_data);
-	lm_element_unref (element);
-
-	g_free (id);
-
-	
-}
-
-#endif
