@@ -1,7 +1,7 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
 /*
+ * Copyright (C) 2003 Imendio HB
  * Copyright (C) 2003 Mikael Hallendal <micke@imendio.com>
- * Copyright (C) 2003 CodeFactory AB. 
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public License as
@@ -27,6 +27,8 @@
 #endif
 
 #include <loudmouth/lm-message.h>
+#include <loudmouth/lm-proxy.h>
+#include <loudmouth/lm-ssl.h>
 
 #define LM_CONNECTION(o) (LmConnection *) o;
 
@@ -56,14 +58,21 @@ typedef enum {
 	LM_DISCONNECT_REASON_UNKNOWN
 } LmDisconnectReason;
 
-typedef void (* LmResultFunction)     (LmConnection *connection,
-				       gboolean      success,
-				       gpointer      user_data);
+typedef enum {
+	LM_CONNECTION_STATE_DISCONNECTED,
+	LM_CONNECTION_STATE_CONNECTING,
+	LM_CONNECTION_STATE_CONNECTED,
+	LM_CONNECTION_STATE_AUTHENTICATING,
+	LM_CONNECTION_STATE_AUTHENTICATED
+} LmConnectionState;
 
-typedef void (* LmDisconnectFunction) (LmConnection *connection,
-				       LmDisconnectReason reason,
-				       gpointer           user_data);
+typedef void          (* LmResultFunction)     (LmConnection       *connection,
+						gboolean            success,
+						gpointer            user_data);
 
+typedef void          (* LmDisconnectFunction) (LmConnection       *connection,
+						LmDisconnectReason  reason,
+						gpointer            user_data);
 
 LmConnection *lm_connection_new               (const gchar        *server);
 gboolean      lm_connection_open              (LmConnection       *connection,
@@ -71,9 +80,13 @@ gboolean      lm_connection_open              (LmConnection       *connection,
 					       gpointer            user_data,
 					       GDestroyNotify      notify,
 					       GError            **error);
+
 gboolean      lm_connection_open_and_block    (LmConnection       *connection,
 					       GError            **error);
 
+void          lm_connection_cancel_open        (LmConnection      *connection);
+						
+
 gboolean      lm_connection_close             (LmConnection       *connection,
 					       GError            **error);
 gboolean      lm_connection_authenticate      (LmConnection       *connection,
@@ -100,11 +113,17 @@ void          lm_connection_set_server        (LmConnection       *connection,
 guint         lm_connection_get_port          (LmConnection       *connection);
 void          lm_connection_set_port          (LmConnection       *connection,
 					       guint               port);
-gboolean      lm_connection_supports_ssl      (void);
-gboolean      lm_connection_get_use_ssl       (LmConnection       *connection);
-void          lm_connection_set_use_ssl       (LmConnection       *connection,
-					       gboolean            use_ssl);
-					       
+
+LmSSL *       lm_connection_get_ssl           (LmConnection       *connection);
+void          lm_connection_set_ssl           (LmConnection       *connection,
+					       LmSSL              *ssl);
+/*
+ * These will most likely change in the API 
+ *
+LmProxy *     lm_connection_get_proxy         (LmConnection       *connection);
+void          lm_connection_set_proxy         (LmConnection       *connection,
+					       LmProxy            *proxy);
+*/
 gboolean      lm_connection_send              (LmConnection       *connection,
 					       LmMessage          *message,
 					       GError            **error);
@@ -134,6 +153,7 @@ lm_connection_set_disconnect_function         (LmConnection       *connection,
 gboolean      lm_connection_send_raw          (LmConnection       *connection,
 					       const gchar        *str,
 					       GError            **error);
+LmConnectionState lm_connection_get_state     (LmConnection       *connection);
 LmConnection* lm_connection_ref               (LmConnection       *connection);
 void          lm_connection_unref             (LmConnection       *connection);
 
